# autopep8: off
import sys; sys.dont_write_bytecode = True
# autopep8: on
from .antidb import *

__version__ = 'v1.2.0'
__authors__ = [{'name': 'Platon Bykadorov',
                'email': 'platon.work@gmail.com',
                'years': '2023'}]
